     /**
      * \brief JSON : Utilities
      *
      * Utilities module for the JSON service program.
      *
      * \author Mihael Schmidt
      * \date   31.01.2010
      *
      * \link http://www.rpgnextgen.com RPG Next Gen
      * \link http://www.json.org JSON
      *
      * \todo json_util_dataStructureToJsonObject : pointer to datastructure as parm
      *
      * \rev 01.08.2010 Mihael Schmidt
      *      added date data type support
      *
      * \rev 26.11.2010 Mihael Schmidt
      *      added toFile/fromFile procedures
      */

      *-------------------------------------------------------------------------
      * Copyright (c) 2010 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the JSON service program.
      *
      * JSON service program is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * JSON service program is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with JSON service program. If not, see http://www.gnu.org/licenses/.
      *-------------------------------------------------------------------------


     H nomain
     H bnddir('QC2LE')
     H copyright('Copyright (c) 2010 Mihael Schmidt. All rights reserved.')


      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D errno           PR              *   extproc('__errno')
      *
     D strerr          PR              *   extproc('strerror')
     D    errnum                     10I 0 value
      *
      /copy 'json_h.rpgle'
      /copy 'json_int_h.rpgle'
      /copy 'ifsio_h.rpgle'
      /copy 'libc_h.rpgle'
      /copy 'ceeapi_h.rpgle'
      /copy 'llist_h.rpgle'

      /if not defined (QUSEC)
      /define QUSEC
      /copy QSYSINC/QRPGLESRC,QUSEC
      /endif


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
      /copy 'json_c.rpgle'


      *-------------------------------------------------------------------------
      * Global Variables
      *-------------------------------------------------------------------------
      *
      * Parameters passed to CEEDOD
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------

     /**
      * \brief Get date entry
      *
      * A convenience procedure for getting a date from a string entry.
      * The string must be in *ISO format (yyyy-mm-dd).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a date string
      * in *ISO format an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   01.08.2010
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Date
      */
     P json_util_getDate...
     P                 B                   export
     D                 PI              D   opdesc
     D   json                          *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(json)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)

     D retVal          S               D
      /free
       isJSONImpl(json);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(json : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %date(%str(entry.value) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Add date entry
      *
      * Convenience procedure for adding a date string in *ISO format to the
      * json object.
      *
      * \author Mihael Schmidt
      * \date 2010-08-01
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Date
      */
     P json_util_putDate...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         D   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S             10A
     D keyLength       S             10I 0
     D hexnull         S              1A   inz(x'00')
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       tmpValue = %char(value : *ISO);

       // check if key is already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue) + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : %size(tmpValue));

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + %size(tmpValue) : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get date entry
      *
      * A convenience procedure for getting a date from a string entry.
      * The string must be in *ISO format (yyyy-mm-dd).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a date string
      * in *ISO format an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   01.08.2010
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Date
      */
     P jsona_util_getDate...
     P                 B                   export
     D                 PI              D
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = list_get(jsonArrPtr : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           return %date(%str(ptr + 4) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return %date(); // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Add date entry
      *
      * Convenience procedure for adding a date string in *ISO format to the
      * JSON array.
      *
      * \author Mihael Schmidt
      * \date   01.08.2010
      *
      * \param Pointer to the JSON array
      * \param Date
      * \param Index (default: last)
      */
     P jsona_util_putDate...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                         D   const
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry) inz
      /free
       entry.type = JSON_TYPE_STRING;
       entry.string = %char(value : *ISO);

       if (%parms() = 2);
         list_add(jsonArrPtr : %addr(entry) : 10 + 4);
       else;
         list_add(jsonArrPtr : %addr(entry) : 10 + 4 : index);
       endif;
      /end-free
     P                 E



     /**
      * \brief Output JSON Object to stream file
      *
      * \param JSON Object
      * \param Stream file path
      * \param CCSID (optional)
      * \param Delete if file exist (optional)
      *
      * \throws CPF9898 Error on opening file
      */
     P json_util_toFile...
     P                 B                   export
     D                 PI
     D  json                           *   const
     D  filePath                   1024A   const
     D  pCcsid                       10I 0 const options(*nopass : *omit)
     D  pDeleteFile                    N   const options(*nopass : *omit)
     D  pOutputFormat...
     D                               10I 0 const options(*nopass)
     D  pPrettyPrintFormat...
     D                                     likeds(json_prettyPrintFormat)
     D                                     const options(*nopass)
      *
     D outputFormat    S             10I 0
     D prettyPrintOption...
     D                 S               N   inz(*off)
     D prettyPrintFormat...
     D                 DS                  likeds(json_prettyPrintFormat) inz
     D ccsid           S             10I 0 inz(0)
     D deleteFile      S               N   inz(*off)
     D jsonString      S               *
     D fileHandle      S             10I 0
     D flags           S             10I 0
     D mode            S             10I 0
     D errorNumber     S             10I 0 based(errorPtr)
     D errorString     S           1024A   varying
     D msgdata         S            512A
     D msgkey          S              4A

      /free
       if (%parms() >= 3 and %addr(pCcsid) <> *null);
         ccsid = pCcsid;
       endif;

       if (%parms() >= 4 and %addr(pDeleteFile) <> *null);
         deleteFile = pDeleteFile;
       endif;

       if (%parms() >= 5);
         outputFormat = pOutputFormat;
         prettyPrintFormat.useTabs = *off;
         prettyPrintFormat.numberWhitespace = 4;
         prettyPrintFormat.breakAfterElement = *on;
         prettyPrintFormat.breakAfterObjectBrace = *on;
         prettyPrintFormat.breakBeforeArrayBrace = *on;
         prettyPrintFormat.breakAfterArrayBrace = *on;
       endif;

       if (%parms() >= 6);
         prettyPrintFormat = pPrettyPrintFormat;
       endif;

       if (outputFormat = JSON_OUTPUT_PRETTY_PRINT);
         prettyPrintOption = *on;
       endif;


       if (deleteFile);
         if (access(%trimr(filePath) : F_OK) >= 0);
           // file exists => delete
           system('RMVLNK OBJLNK(''' + %trimr(filePath) + ''')');
         endif;
       endif;


       // create file
       flags = O_WRONLY + O_CREAT + O_APPEND + O_TEXTDATA +
               O_CCSID + O_TEXT_CREAT;
       mode =  S_IRUSR + S_IWUSR + S_IRGRP + S_IROTH;
       fileHandle = open(%trimr(filePath) : flags : mode : ccsid : 0);
       if (fileHandle < 0);
         // could not create file
         errorPtr = errno();
         errorString = 'I/O Error: ' + %str(strerr(errorNumber));
         msgdata = 'Could not create file ' + %trimr(filePath) +
                   ': ' + errorString;
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       jsonString = json_toString(json  : outputFormat : prettyPrintFormat);

       callp write(fileHandle : jsonString : strlen(jsonString));

       dealloc(n) jsonString;

       callp close(fileHandle);
      /end-free
     P                 E


     /**
      * \brief Read JSON from stream file
      *
      * \param Stream file path
      * \param Delete file after reading (optional)
      *
      * \return JSON object
      *
      * \throws CPF9898 Error on reading file
      * \throws CPF9898 File not found
      * \throws CPF9898 Error buildng JSON object
      */
     P json_util_fromFile...
     P                 B                   export
     D                 PI              *
     D  filePath                   1024A   const
     D  pDeleteFile                    N   const options(*nopass)
      *
     D json            S               *
     D deleteFile      S               N   inz(*off)
     D buffer          S               *
     D fileSize        S             10I 0
     D fileStatus      DS                  likeds(statds)
     D fileHandle      S             10I 0
     D flags           S             10I 0
     D errorNumber     S             10I 0 based(errorPtr)
     D errorString     S           1024A   varying
     D msgdata         S            512A
     D msgkey          S              4A

      /free
       if (%parms() = 2);
         deleteFile = pDeleteFile;
       endif;

       if (stat(%trimr(filePath) : %addr(fileStatus)) < 0);
         msgdata = 'Could not query file attributes for file ' +
                   %trimr(filepath) + '.';
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       flags = O_RDONLY + O_TEXTDATA;
       fileHandle = open(%trimr(filePath) : flags);
       if (fileHandle < 0);
         // could not open file
         errorPtr = errno();
         errorString = 'I/O Error: ' + %str(strerr(errorNumber));
         msgdata = 'Could not open file ' + %trimr(filePath) + ': '+errorString;
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       buffer = %alloc(fileStatus.st_size);

       callp read(fileHandle : buffer : fileStatus.st_size);

       callp close(fileHandle);

       json = json_parse(buffer);

       // clean up
       dealloc buffer;

       if (deleteFile);
         if (access(%trimr(filePath) : F_OK) >= 0);
           // file exists => delete
           system('RMVLNK OBJLNK(''' + %trimr(filePath) + ''')');
         endif;
       endif;

       return json;
      /end-free
     P                 E


     /**
      * \brief Get decimal entry
      *
      * A convenience procedure for getting a decimal value from a string entry.
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a number
      * an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   09.04.2011
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Decimal value
      */
     P json_util_getDecimal...
     P                 B                   export
     D                 PI            30P10 opdesc
     D   json                          *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(json)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)

     D retVal          S             30P10
      /free
       isJSONImpl(json);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(json : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %dec(%str(entry.value) : 30 : 10);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E
    
    
     /**
      * \brief Get decimal entry
      *
      * A convenience procedure for getting a decimal value from a string entry.
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a number
      * an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   09.04.2011
      *
      * \param Pointer to the JSON array 
      * \param Index
      *
      * \return Decimal value
      */
     P jsona_util_getDecimal...
     P                 B                   export
     D                 PI            30P10
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
     D retVal          S             30P10
      /free
       ptr = list_get(jsonArrPtr : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %dec(%str(ptr + 4) : 30 : 10);
           return retVal;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return 0; // dummy => will never come to this point
      /end-free
     P                 E
     
     
     /**
      * \brief Add decimal entry
      *
      * Convenience procedure for adding a decimal value as a string to the 
      * json object.
      *
      * \author Mihael Schmidt
      * \date   09.04.2011
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Decimal value
      * \param Decimal positions (default: 2)
      */
     P json_util_putDecimal...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                       30P10 const
     D   pDecimalPositions...
     D                                2P 0 const options(*nopass)
      *
     D decimalPositions...
     D                 S              2P 0 inz(2)
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S             32A
     D tmpValueLength  S             10I 0
     D keyLength       S             10I 0
     D hexnull         S              1A   inz(x'00')
     D decimalSepIndex...
     D                 S             10I 0
      /free
       isJSONImpl(jsonPtr);

       if (%parms() = 4);
         decimalPositions = pDecimalPositions;
       endif;

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       tmpValue = %char(value);

       // max. decimal precision of 10
       if (decimalPositions > 10);
         decimalPositions = 10;
       elseif (decimalPositions < 0); // min. decimal precision of 0
         decimalPositions = 0;
       endif;

       // limit the string to the passed number of decimal positions
       decimalSepIndex = %scan('.' : tmpValue);
       if (decimalSepIndex = 0);
         decimalSepIndex = %scan(',' : tmpValue);
       endif;
       
       if (decimalSepIndex <> 0);
         tmpValueLength = decimalSepIndex + decimalPositions; 
       else;
         tmpValueLength = %len(%trim(tmpvalue));
       endif;

       // check if key is already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = tmpValueLength;
         newEntry.value = %alloc(tmpValueLength + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = tmpValueLength;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : tmpValueLength);

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + tmpValueLength : %addr(hexNull) : 1);
      /end-free
     P                 E
    
    
     /**
      * \brief Add decimal entry
      *
      * Convenience procedure for adding a decimal value as a string to the
      * JSON array.
      *
      * \author Mihael Schmidt
      * \date   12.04.2011
      *
      * \param Pointer to the JSON array
      * \param Decimal value
      * \param Decimal positions (default: 2)
      * \param Index (default: last)
      */
     P jsona_util_putDecimal...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                       30P10 const
     D   pDecimalPositions...
     D                                2P 0 const options(*nopass : *omit)
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry) inz
     D decimalSepIndex...
     D                 S             10I 0
     D decimalPositions...
     D                 S              2P 0 inz(2)
     D tmpValueLength  S             10I 0
      /free
       entry.type = JSON_TYPE_STRING;
       entry.string = %char(value);
       
       if (%parms() >= 3 and %addr(pDecimalPositions) <> *null);
         decimalPositions = pDecimalPositions;
       endif;

       // max. decimal precision of 10
       if (decimalPositions > 10);
         decimalPositions = 10;
       elseif (decimalPositions < 0); // min. decimal precision of 0
         decimalPositions = 0;
       endif;
       
       // limit the string to the passed number of decimal positions
       decimalSepIndex = %scan('.' : entry.string);
       if (decimalSepIndex = 0);
         decimalSepIndex = %scan(',' : entry.string);
       endif;
       
       if (decimalSepIndex <> 0);
         tmpValueLength = decimalSepIndex + decimalPositions; 
       else;
         tmpValueLength = %len(%trim(entry.string));
       endif;

       if (%parms() < 4);
         list_add(jsonArrPtr : %addr(entry) : tmpValueLength + 4);
       else;
         list_add(jsonArrPtr : %addr(entry) : tmpValueLength + 4 : index);
       endif;
      /end-free
     P                 E